// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#include <iostream>

#include <fmt/core.h>

#include <amdis/AMDiS.hpp>
#include <amdis/ProblemStat.hpp>
#include <amdis/Operators.hpp>
#include <amdis/common/Literals.hpp>
#include <amdis/gridfunctions/Integrate.hpp>

using namespace AMDiS;
using namespace Dune::Indices;

// 1 component with polynomial degree 1
using Param   = YaspGridBasis<AMDIS_DIM, 1>;
using ElliptProblem = ProblemStat<Param>;

int main(int argc, char** argv)
{
  AMDiS::init(argc, argv);

  int numLevels = AMDIS_DIM == 2 ? 8 : 5;
  if (argc > 2)
    numLevels = std::atoi(argv[2]);

  auto f = [](auto const& x) {
    double r2 = dot(x,x);
    double ux = std::exp(-10.0 * r2);
    return -(400.0 * r2 - 20.0 * x.size()) * ux;
  };
  auto g = [](auto const& x){ return std::exp(-10.0 * dot(x,x)); };
  auto grad_g = [](auto const& x) -> FieldMatrix<double,1,AMDIS_DIM> {
    return {-20.0 * std::exp(-10.0 * dot(x,x)) * x};
  };

  using Grid = Dune::YaspGrid<AMDIS_DIM>;
  auto gridPtr = MeshCreator<Grid>::create("mesh");
  auto basis = makeBasis(gridPtr->leafGridView(), lagrange<1>());

  auto prob = makeProblemStat("ellipt", *gridPtr, basis);
  prob.initialize(INIT_ALL);

  auto opL = makeOperator(tag::gradtest_gradtrial{}, 1.0);
  prob.addMatrixOperator(opL);

  auto opForce = makeOperator(tag::test{}, f, 6);
  prob.addVectorOperator(opForce, _0);

  // set boundary condition
  auto boundary = [](auto const& x){ return x[0] < 1.e-8 || x[1] < 1.e-8 || x[0] > 1.0-1.e-8 || x[1] > 1.0-1.e-8; };
  prob.addDirichletBC(boundary, _0, _0, g);

  AdaptInfo adaptInfo("adapt");

  std::vector<double> errL2; errL2.reserve(numLevels);
  std::vector<double> errH1; errH1.reserve(numLevels);
  std::vector<double> widths; widths.reserve(numLevels);
  for (int i = 0; i < numLevels; ++i) {
    prob.getGrid()->globalRefine(1);
    auto gridView = prob.getGlobalBasis()->gridView();

    double h = 0;
    for (auto const& e : edges(gridView))
      h = std::max(h, e.geometry().volume());
    widths.push_back(h);

    prob.buildAfterCoarsen(adaptInfo, Flag(0));
    prob.solve(adaptInfo);

    double errorL2 = integrate(sqr(g - prob.getSolution(_0)), gridView, 6);
    errL2.push_back(std::sqrt(errorL2));
    double errorH1 = errorL2 + integrate(unary_dot(grad_g - gradientAtQP(prob.getSolution(_0))), gridView, 6);
    errH1.push_back(std::sqrt(errorH1));

#if WRITE_FILES
    Dune::VTKWriter<typename ElliptProblem::GridView> vtkWriter(prob.getGlobalBasis()->gridView());
    vtkWriter.addVertexData(prob.getSolution(_0), Dune::VTK::FieldInfo("u", Dune::VTK::FieldInfo::Type::scalar, 1));
    vtkWriter.write("u_" + std::to_string(i));
#endif
  }

  std::cout << "\n";
  fmt::print("{:5} | {:12} | {:12} | {:12} | {:12} | {:12}\n",
             "level", "h", "|u - u*|_L2","|u - u*|_H1","eoc_L2","eoc_H1");
  fmt::print("{0:->7}{0:->15}{0:->15}{0:->15}{0:->15}{1:->14}","+","\n");

  std::vector<double> eocL2(numLevels, 0.0), eocH1(numLevels, 0.0);
  for (int i = 1; i < numLevels; ++i) {
    eocL2[i] = std::log(errL2[i]/errL2[i-1]) / std::log(widths[i]/widths[i-1]);
    eocH1[i] = std::log(errH1[i]/errH1[i-1]) / std::log(widths[i]/widths[i-1]);
  }

  for (int i = 0; i < numLevels; ++i)
    fmt::print("{:<5} | {:<12} | {:<12} | {:<12} | {:<12} | {:<12}\n",
                i+1, widths[i], errL2[i], errH1[i], eocL2[i], eocH1[i]);

  AMDiS::finalize();
  return 0;
}
