#include <iostream>
#include <ctime>
#include <cmath>

#include <amdis/AMDiS.hpp>
#include <amdis/ProblemStat.hpp>
#include <amdis/Operators.hpp>

#ifdef DOW
#undef DOW
#endif
#define DOW AMDIS_DOW

using namespace AMDiS;

int main(int argc, char** argv)
{
    AMDiS::init(argc, argv);

    using Grid = Dune::YaspGrid<AMDIS_DIM>;
    auto gridPtr = MeshCreator<Grid>::create("mesh");

    auto preBasis = compose(repeat<AMDIS_DOW>(lagrange<2>()), lagrange<1>());
    auto basis = makeBasis(gridPtr->leafGridView(), preBasis);

    auto prob = makeProblemStat("stokes", *gridPtr, basis);
    prob.initialize(INIT_ALL);

    double viscosity = 1.0;
    Parameters::get("stokes->viscosity", viscosity);

    // tree-paths for components
    auto _v = 0_c;
    auto _p = 1_c;

    // <viscosity*grad(u_i), grad(v_i)>
    for (std::size_t i = 0; i < DOW; ++i) {
      auto opL = makeOperator(tag::gradtest_gradtrial{}, viscosity);
      prob.addMatrixOperator(opL, treepath(_v,i), treepath(_v,i));
    }

    // <d_i(v_i), p>
    auto opP = makeOperator(tag::divtestvec_trial{}, 1.0);
    prob.addMatrixOperator(opP, _v, _p);

    // <q, d_i(u_i)>
    auto opDiv = makeOperator(tag::test_divtrialvec{}, 1.0);
    prob.addMatrixOperator(opDiv, _p, _v);

    auto opZero = makeOperator(tag::test_trial{}, 0.0);
    prob.addMatrixOperator(opZero, _p, _p);


    // define boundary regions
    auto left = [](auto const& x) { return x[0] < 1.e-8; };
    auto not_left = [](auto const& x) { return x[0] > 1.0 - 1.e-8 || x[1] < 1.e-8 || x[1] > 1.0 - 1.e-8; };

    // define boundary values
    auto parabolic_y = [](auto const& x) -> Dune::FieldVector<double,DOW>
    {
      return {0.0, x[1]*(1.0 - x[1])};
    };

    auto zero = [](auto const& x) -> Dune::FieldVector<double,DOW>
    {
      return {0.0, 0.0};
    };

    // set boundary conditions for velocity
    prob.addDirichletBC(left, _v, _v, parabolic_y);
    prob.addDirichletBC(not_left, _v, _v, zero);

    prob.addDirichletBC([](auto const& x) { return x[0] < 1.e-8 && x[1] < 1.e-8; }, _p, _p, 0.0);

    AdaptInfo adaptInfo("adapt");

    // assemble and solve system
    prob.buildAfterCoarsen(adaptInfo, Flag(0));

    // write matrix to file
    mtl::io::matrix_market_ostream out("matrix_stokes1.mtx");
    out << prob.getSystemMatrix()->getMatrix();
    std::cout << prob.getSystemMatrix()->getMatrix() << '\n';

    prob.solve(adaptInfo);

    // output solution
    prob.writeFiles(adaptInfo);

    AMDiS::finalize();
    return 0;
}
